using System;
using System.Collections.Generic;
using System.Text;

namespace Chessticles
{
	public class Game
	{
		public enum EState
		{
			NotStarted,
			WaitingMove,
			Finished,
		}

		public enum EPlayerType
		{
			Human,
			AI,
		}

		#region Properties

		public EState State
		{
			get { return m_state; }
		}

		public EPlayer CurrentPlayer
		{
			get { return m_playerTurn; }
		}

		public EWinner Winner
		{
			get { return m_winner; }
		}

		public int MoveCount
		{
			get { return m_iMoveCount; }
		}

		public int NumRows
		{
			get { return GameModel.NUM_ROWS; }
		}

		public int NumCols
		{
			get { return GameModel.NUM_COLS; }
		}

		#endregion // Properties


		public Game() 
		{
			m_model = new GameModel();
			m_players = new SPlayer[2];
			m_players[(int)EPlayer.White] = new SPlayer(EPlayerType.Human, null); ;
			m_players[(int)EPlayer.Black] = new SPlayer(EPlayerType.Human, null); ;

			ResetGame();
		}

		public void SetPlayerType(EPlayer player, EPlayerType eType)
		{
			if (player != EPlayer.None)
			{
				m_players[(int)player].type = eType;
			}
		}

		public EPlayerType GetPlayerType(EPlayer player) 
		{
			return (player != EPlayer.None ? m_players[(int)player].type : EPlayerType.Human);
		}

		public bool IsPlayerAI(EPlayer player)
		{
			return (player != EPlayer.None && m_players[(int)player].type == EPlayerType.AI);
		}

		public bool AllPlayersAI()
		{
			foreach (SPlayer player in m_players)
			{
				if (player.type == EPlayerType.Human)
				{
					return false;
				}
			}

			return true;
		}

		public bool InCheck(EPlayer player)
		{
			return m_model.InCheck(player);
		}

		public void Start()
		{
			ResetGame();
			m_model.StartGame();

			for (int i = 0; i < m_players.Length; ++i )
			{
				if (m_players[i].type == EPlayerType.AI)
				{
					m_players[i].ai = new AIPlayerRandom(this, (EPlayer)i);
				}
			}

			m_state = EState.WaitingMove;
			m_playerTurn = EPlayer.White;
		}

		public void UpdateGame()
		{
			if (m_state == EState.WaitingMove)
			{
				if (IsPlayerAI(m_playerTurn))
				{
					DoAIMove();
				}

				ProcessMove();
			}
		}

		public EPlayer GetPositionOwner(Position pos) 
		{
			return m_model.GetPositionOwner(pos);
		}

		public EPieceType GetPieceType(Position pos)
		{
			return m_model.GetPieceType(pos);
		}

		public IPiece GetPiece(Position pos)
		{
			return m_model.GetPiece(pos);
		}

		public List<IPiece> GetCurrentPieces(EPlayer player)
		{
			return m_model.GetCurrentPieces(player);
		}

		public void RequestMove(SMove move)
		{
			// register movement request as long as the piece being
			// moved belongs to the current player
			if (move.piece.Owner == m_playerTurn)
			{
				m_moveRequest = move;
			}
		}

		public void ResetGame()
		{
			m_state = EState.NotStarted;
			m_winner = EWinner.Unfinished;
			m_playerTurn = EPlayer.None;
			m_iMoveCount = 0;
			m_moveRequest.type = EMoveType.Invalid;

			m_model.ResetGame();
		}


		private void ProcessMove()
		{
			if (m_moveRequest.type != EMoveType.Invalid && m_model.ProcessMove(ref m_moveRequest))
			{
				++m_iMoveCount;
				m_playerTurn = (m_playerTurn == EPlayer.White) ? EPlayer.Black : EPlayer.White;
						
				m_winner = m_model.TestForWinCondition(m_playerTurn);
				switch (m_winner)
				{
					case EWinner.Draw:
					case EWinner.White:
					case EWinner.Black:
						m_state = EState.Finished;
						break;
					case EWinner.Unfinished:
					default:
						m_state = EState.WaitingMove;
						break;
				}
			}		
		}

		private void DoAIMove()
		{
			if (m_players[(int)m_playerTurn].ai != null)
			{
				m_players[(int)m_playerTurn].ai.DoMove();
			}
		}

		
		private struct SPlayer
		{
			public EPlayerType type;
			public IAIPlayer ai;

			public SPlayer(EPlayerType _type, IAIPlayer _ai)
			{
				type = _type;
				ai = _ai;
			}
		}


		#region Members

		private GameModel m_model;

		private EState m_state;
		private EWinner m_winner;
		private EPlayer m_playerTurn;

		private SPlayer[] m_players;
		private SMove m_moveRequest;
		private int m_iMoveCount;

		#endregion // Members
	}
}
