using System;
using System.Collections.Generic;
using System.Text;

namespace Chessticles
{
	public enum EWinner
	{
		Unfinished, // game not finished yet
		Draw,
		White,
		Black,
	}

	public enum EPlayer
	{
		None = -1,
		White = 0,
		Black = 1,
	}

	public enum EMoveType
	{
		Invalid,
		Standard,
		EnPassant,
		PawnPromotion,
		Castling,
	}

	public struct SMove
	{
		public IPiece piece;
		public EMoveType type;
		public Position newPos;

		public IPiece target; // for En Passant or Castling moves
		public Position rookPos; // for Castling moves
		public EPieceType promotionType; // for Pawn Promotion

		public SMove(IPiece _piece, Position _newPos)
		{
			type = EMoveType.Standard;
			piece = _piece;
			newPos = _newPos;

			target = null;
			rookPos = null;
			promotionType = EPieceType.None;
		}

		public SMove(IPiece _piece, Position _newPos, IPiece targetPiece)
		{
			type = EMoveType.EnPassant;
			piece = _piece;
			newPos = _newPos;
			target = targetPiece;

			rookPos = null;
			promotionType = EPieceType.None;
		}

		public SMove(IPiece _piece, Position _newPos, EPieceType _promotionType)
		{
			type = EMoveType.PawnPromotion;
			piece = _piece;
			newPos = _newPos;
			promotionType = _promotionType;

			target = null;
			rookPos = null;
		}

		public SMove(IPiece _piece, Position _newPos, IPiece rook, Position newRookPos)
		{
			type = EMoveType.Castling;
			piece = _piece;
			newPos = _newPos;
			target = rook;
			rookPos = newRookPos;

			promotionType = EPieceType.None;
		}

	};

	public class GameModel
	{
		public const int NUM_ROWS = 8;
		public const int NUM_COLS = 8;
		public const int NUM_PLAYERS = 2;
		
		#region Properties

		public SMove LastMove
		{
			get { return m_lastMove; }
		}

		public bool TestingForCheck
		{
			get { return m_bTestingForCheck; }
		}

		#endregion // Properties


		public GameModel()
		{
			ResetGame();
		}

		public void ResetGame()
		{
			// Create the pieces
			m_iActivePieces = 0;
			m_pieces = new Dictionary<EPlayer, List<Piece> >();
			m_pieces[EPlayer.White] = new List<Piece>();
			m_pieces[EPlayer.Black] = new List<Piece>();
			InitMainPieceRow(0, EPlayer.Black);
			InitPawnRow(1, EPlayer.Black);
			InitMainPieceRow(NUM_ROWS - 1, EPlayer.White);
			InitPawnRow(NUM_ROWS - 2, EPlayer.White);

			// Initialize the board
			m_board = new Piece[NUM_ROWS, NUM_COLS];
			foreach (List<Piece> pieces in m_pieces.Values)
			{
				foreach (Piece piece in pieces)
				{
					m_board[piece.Position.row, piece.Position.col] = piece;
				}
			}
			
			// Initialize 'in check' dictionary
			m_inCheck = new Dictionary<EPlayer, bool>();
			m_inCheck[EPlayer.White] = false;
			m_inCheck[EPlayer.Black] = false;

			m_lastMove.type = EMoveType.Invalid;
			m_bTestingForCheck = false;
		}

		public void StartGame()
		{
			FullUpdateValidMoves(EPlayer.White);
		}

		public EPlayer GetPositionOwner(Position pos)
		{
			Piece piece = pos.IsValid() ? m_board[pos.row, pos.col] : null;
			return (piece != null ? piece.Owner : EPlayer.None);
		}

		public EPieceType GetPieceType(Position pos)
		{
			Piece piece = pos.IsValid() ? m_board[pos.row, pos.col] : null;
			return (piece != null ? piece.Type : EPieceType.None);
		}

		public IPiece GetPiece(Position pos)
		{
			return (pos.IsValid() ? (IPiece)m_board[pos.row, pos.col] : null);
		}

		public List<IPiece> GetCurrentPieces(EPlayer player)
		{
			List<IPiece> pieces = new List<IPiece>(m_pieces[player].Count);
			foreach (Piece piece in m_pieces[player])
			{
				if (piece.Owner == player && piece.OnBoard)
				{
					pieces.Add((IPiece)piece);
				}
			}

			return pieces;
		}

		public bool InCheck(EPlayer player)
		{
			return (m_inCheck.ContainsKey(player) ? m_inCheck[player] : false);
		}

		public bool ProcessMove(ref SMove move)
		{
			Piece piece = (Piece)move.piece;
			switch (move.type)
			{
				case EMoveType.PawnPromotion:
					{	// Same as a standard move, but with a piece type change added
						if (piece.Type == EPieceType.Pawn)
						{
							piece = DoPawnPromotion((Pawn)piece, move.promotionType);
						}

						goto case EMoveType.Standard;
					}
				case EMoveType.Standard:
					{
						// Remove any captured piece
						Piece captured = (move.newPos.IsValid()) ? m_board[move.newPos.row, move.newPos.col] : null;
						if (captured != null)
						{
							captured.OnBoard = false;
							m_iActivePieces--;
						}

						// Move to new position
						m_board[piece.Position.row, piece.Position.col] = null;
						m_board[move.newPos.row, move.newPos.col] = piece;
						piece.Position = move.newPos;

						break;
					}
				case EMoveType.EnPassant:
					{	// Remove captured pawn
						Piece captured = (Piece)move.target;
						m_board[captured.Position.row, captured.Position.col] = null;
						captured.OnBoard = false;
						m_iActivePieces--;

						// Move to new position
						m_board[piece.Position.row, piece.Position.col] = null;
						m_board[move.newPos.row, move.newPos.col] = piece;
						piece.Position = move.newPos;

						break;
					}
				case EMoveType.Castling:
					{	// Move King
						m_board[piece.Position.row, piece.Position.col] = null;
						m_board[move.newPos.row, move.newPos.col] = piece;
						piece.Position = move.newPos;

						// Move Rook
						Piece rook = (Piece)move.target;
						m_board[move.target.Position.row, move.target.Position.col] = null;
						m_board[move.rookPos.row, move.rookPos.col] = rook;
						rook.Position = move.rookPos;

						break;
					}
				case EMoveType.Invalid:
				default:
					// Should never get here
					Console.WriteLine("Invalid move received");
					return false;
			}

			m_lastMove = move;

			FullUpdateValidMoves(piece.Owner);

			return true;
		}

		private Piece DoPawnPromotion(Pawn piece, EPieceType type)
		{
			Piece newPiece = null;

			switch (type)
			{
				case EPieceType.Bishop:
					newPiece = new Bishop(piece);
					break;
				case EPieceType.Knight:
					newPiece = new Knight(piece);
					break;
				case EPieceType.Queen:
					newPiece = new Queen(piece);
					break;
				case EPieceType.Rook:
					newPiece = new Rook(piece);
					break;
				default:
					// Should never get here
					Console.WriteLine("Got pawn promotion request for type {0}", type);
					return null;
			}

			// Replace old piece on board and pieces list with new one
			m_board[piece.Position.row, piece.Position.col] = newPiece;
			m_pieces[newPiece.Owner].Remove(piece);
			m_pieces[newPiece.Owner].Add(newPiece);

			return newPiece;
		}

		private void InitMainPieceRow(int row, EPlayer player)
		{
			m_pieces[player].Add(new Rook  (this, player, new Position(row, 0)));
			m_pieces[player].Add(new Knight(this, player, new Position(row, 1)));
			m_pieces[player].Add(new Bishop(this, player, new Position(row, 2)));
			m_pieces[player].Add(new Queen(this, player, new Position(row, 3)));
			m_pieces[player].Add(new King(this, player, new Position(row, 4)));
			m_pieces[player].Add(new Bishop(this, player, new Position(row, 5)));
			m_pieces[player].Add(new Knight(this, player, new Position(row, 6)));
			m_pieces[player].Add(new Rook(this, player, new Position(row, 7)));
			m_iActivePieces += 8;
		}

		private void InitPawnRow(int row, EPlayer player)
		{
			for (int i = 0; i < NUM_COLS; ++i)
			{
				m_pieces[player].Add(new Pawn(this, player, new Position(row, i)));
			}
			m_iActivePieces += NUM_COLS;
		}

		public EWinner TestForWinCondition(EPlayer nextPlayer)
		{
			if (!AnyValidMoves(nextPlayer))
			{
				if (m_inCheck[nextPlayer])
				{
					return (nextPlayer == EPlayer.White) ? EWinner.Black : EWinner.White;
				}
				else
				{
					return EWinner.Draw;
				}
			}
			else if (m_iActivePieces == 2)
			{
				// Only kings are left, winning is impossible
				return EWinner.Draw;
			}
						
			return EWinner.Unfinished;
		}

		public bool AnyValidMoves(EPlayer nextPlayer)
		{
			foreach (Piece piece in m_pieces[nextPlayer])
			{
				if (piece.OnBoard && piece.ValidMoves.Count > 0)
				{
					return true;
				}
			}

			return false;
		}

		public bool TestIfMoveCausesSelfCheck(SMove move)
		{
			m_bTestingForCheck = true;

			Piece piece = (Piece)move.piece;
			EPlayer opponent = (piece.Owner == EPlayer.White) ? EPlayer.Black : EPlayer.White;


			// Backup important info
			Position originalPos = move.piece.Position;
			Piece captured = m_board[move.newPos.row, move.newPos.col];
			Piece enPassantCaptured = (move.type == EMoveType.EnPassant) ? (Piece)move.target : null;
			Piece castlingRook = (move.type == EMoveType.Castling) ? (Piece)move.target : null;
			Position castlingRookPos = (castlingRook != null) ? move.target.Position : new Position();
			

			// Make the move
			m_board[piece.Position.row, piece.Position.col] = null;
			m_board[move.newPos.row, move.newPos.col] = piece;
			piece.Position = move.newPos;
			if (captured != null)
			{
				captured.OnBoard = false;
			}
			else if (enPassantCaptured != null)
			{
				enPassantCaptured.OnBoard = false;
				m_board[enPassantCaptured.Position.row, enPassantCaptured.Position.col] = null;
			}
			else if (castlingRook != null)
			{
				m_board[castlingRook.Position.row, castlingRook.Position.col] = null;
				m_board[move.rookPos.row, move.rookPos.col] = castlingRook;
				castlingRook.Position = move.rookPos;
			}
		

			// Test if opponents could capture King on next move (i.e. player is in check with this move)
			bool bInCheck = DoCheckTest(opponent);

			// Second test for castling moves, testing the square the King must move through
			if (!bInCheck && move.type == EMoveType.Castling)
			{
				bInCheck = CastlingSecondaryCheckTest(move, piece, originalPos, opponent);
			}


			// Restore to original state
			m_board[move.newPos.row, move.newPos.col] = captured;
			m_board[originalPos.row, originalPos.col] = piece;
			piece.Position = originalPos;
			if (captured != null)
			{
				captured.OnBoard = true;
			}
			else if (enPassantCaptured != null)
			{
				enPassantCaptured.OnBoard = true;
				m_board[enPassantCaptured.Position.row, enPassantCaptured.Position.col] = enPassantCaptured;
			}
			else if (castlingRook != null)
			{
				m_board[move.rookPos.row, move.rookPos.col] = null;
				m_board[castlingRookPos.row, castlingRookPos.col] = castlingRook;
				castlingRook.Position = castlingRookPos;
			}
			
			m_bTestingForCheck = false;
			
			return bInCheck;
		}

		private bool CastlingSecondaryCheckTest(SMove move, Piece king, Position startPos, EPlayer opponent)
		{
			Position currentPos = king.Position;
			Position testPos = new Position(startPos.row, (startPos.col + currentPos.col) / 2);
			
			m_board[king.Position.row, king.Position.col] = null;
			m_board[testPos.row, testPos.col] = king;
			king.Position = testPos;

			bool bInCheck = DoCheckTest(opponent);

			m_board[testPos.row, testPos.col] = null;
			m_board[currentPos.row, currentPos.col] = king;
			king.Position = currentPos;

			return bInCheck;
		}

		private bool DoCheckTest(EPlayer opponent)
		{
			bool bInCheck = false;
			foreach (Piece testPiece in m_pieces[opponent])
			{
				if (testPiece.UpdateMoves(true))
				{
					bInCheck = true;
					break;
				}
			}

			return bInCheck;
		}

		private void FullUpdateValidMoves(EPlayer currentPlayer)
		{
			// Always update the current player first, so Check test
			if (currentPlayer == EPlayer.White)
			{
				UpdateValidMoves(EPlayer.White);
				UpdateValidMoves(EPlayer.Black);
			}
			else
			{
				UpdateValidMoves(EPlayer.Black);
				UpdateValidMoves(EPlayer.White);
			}
		}

		private void UpdateValidMoves(EPlayer player)
		{
			EPlayer opponent = (player == EPlayer.White) ? EPlayer.Black : EPlayer.White;
			bool bInCheck = false;
			foreach (Piece piece in m_pieces[player])
			{
				if (piece.UpdateMoves(false) == true)
				{
					bInCheck = true;
				}
			}

			m_inCheck[opponent] = bInCheck;
		}

		#region Members

		private Dictionary<EPlayer, List<Piece> > m_pieces;
		private Piece[,] m_board;
		private Dictionary<EPlayer, bool> m_inCheck;
		private SMove m_lastMove;
		private bool m_bTestingForCheck;
		private int m_iActivePieces;

		#endregion // Members
	}
}
