using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace Chessticles
{
	public partial class MainForm : Form
	{
		public MainForm()
		{
			m_game = new Game();
			InitializeComponent();
			m_buttons = new Button[,] { {b00, b01, b02, b03, b04, b05, b06, b07},
			   	                        {b10, b11, b12, b13, b14, b15, b16, b17},
				                          {b20, b21, b22, b23, b24, b25, b26, b27},
				                          {b30, b31, b32, b33, b34, b35, b36, b37},
				                          {b40, b41, b42, b43, b44, b45, b46, b47},
				                          {b50, b51, b52, b53, b54, b55, b56, b57},
				                          {b60, b61, b62, b63, b64, b65, b66, b67},
				                          {b70, b71, b72, b73, b74, b75, b76, b77} };

			Reset();
		}

		private void Reset()
		{
			ResetBoardColors();
			
			lblStatus.Text = "";
			lblMoves.Text = "";
			EnableBoardSquares(false);
			m_selectedPiece = null;
			m_inputState = EInputState.None;
		}

		private void ResetBoardColors()
		{
			// Reset board colors
			int nRows = m_game.NumRows;
			int nCols = m_game.NumCols;
			for (int i = 0; i < nRows; ++i)
			{
				for (int j = 0; j < nCols; ++j)
				{
					Color color = ((i + j) % 2) == 0 ? Color.White : Color.Gray;
					Button button = m_buttons[i, j];
					button.BackColor = color;
					button.FlatAppearance.MouseOverBackColor = color;
					button.FlatAppearance.MouseDownBackColor = color;
				}
			}
		}

		private void StartGame()
		{
			// Get the desired player types
			NewGameForm form = new NewGameForm();
			form.Player1Type = (m_game.IsPlayerAI(EPlayer.White) ? Game.EPlayerType.AI : Game.EPlayerType.Human);
			form.Player2Type = (m_game.IsPlayerAI(EPlayer.Black) ? Game.EPlayerType.AI : Game.EPlayerType.Human);
			form.ShowDialog();
			if (form.DialogResult == DialogResult.OK)
			{
				m_game.SetPlayerType(EPlayer.White, form.Player1Type);
				m_game.SetPlayerType(EPlayer.Black, form.Player2Type);
				m_game.Start();
				EnableBoardSquares(true);
				
				UpdateGame();
			}	
		}
		
		private void UpdateGame()
		{
			// Update only once if there is at least one human player,
			// otherwise keep going until game ends
			do 
			{
				m_game.UpdateGame();
		
				DrawGameBoard();
				UpdateMoveCounter();
				UpdateStatus();

				Application.DoEvents();
			} while (m_game.IsPlayerAI(m_game.CurrentPlayer) && m_game.State == Game.EState.WaitingMove);

			if (m_game.State == Game.EState.WaitingMove)
			{
				m_inputState = EInputState.SelectingPiece;
			}
		}

		private void EnableBoardSquares(bool bEnable)
		{
			int nRows = m_game.NumRows;
			int nCols = m_game.NumCols;
			for (int i = 0; i < nRows; ++i)
				for (int j = 0; j < nCols; ++j)
					m_buttons[i, j].Enabled = bEnable;
		}

		private void DrawGameBoard()
		{
			int nRows = m_game.NumRows;
			int nCols = m_game.NumCols;
			for (int i = 0; i < nRows; ++i)
				for (int j = 0; j < nCols; ++j)
					DrawGameBoardSquare(new Position(i, j));
		}

		private void DrawGameBoardSquare(Position pos)
		{
			EPieceType piece = m_game.GetPieceType(pos);
			EPlayer owner = m_game.GetPositionOwner(pos);
			switch (piece)
			{
				case EPieceType.Bishop:
					m_buttons[pos.row, pos.col].BackgroundImage = (owner == EPlayer.White) ? Chessticles.Properties.Resources.white_bishop : Chessticles.Properties.Resources.black_bishop;
					break;
				case EPieceType.King:
					m_buttons[pos.row, pos.col].BackgroundImage = (owner == EPlayer.White) ? Chessticles.Properties.Resources.white_king : Chessticles.Properties.Resources.black_king;
					break;
				case EPieceType.Knight:
					m_buttons[pos.row, pos.col].BackgroundImage = (owner == EPlayer.White) ? Chessticles.Properties.Resources.white_knight : Chessticles.Properties.Resources.black_knight;
					break;
				case EPieceType.Pawn:
					m_buttons[pos.row, pos.col].BackgroundImage = (owner == EPlayer.White) ? Chessticles.Properties.Resources.white_pawn : Chessticles.Properties.Resources.black_pawn;
					break;
				case EPieceType.Queen:
					m_buttons[pos.row, pos.col].BackgroundImage = (owner == EPlayer.White) ? Chessticles.Properties.Resources.white_queen : Chessticles.Properties.Resources.black_queen;
					break;
				case EPieceType.Rook:
					m_buttons[pos.row, pos.col].BackgroundImage = (owner == EPlayer.White) ? Chessticles.Properties.Resources.white_rook : Chessticles.Properties.Resources.black_rook;
					break;
				case EPieceType.None:
					m_buttons[pos.row, pos.col].BackgroundImage = null;
					break;
			}			
		}

		private void UpdateStatus()
		{
			Game.EState eState = m_game.State;
			switch (eState)
			{
				case Game.EState.Finished:
					DoFinished();
					break;
				case Game.EState.WaitingMove:
					lblStatus.Text = (m_game.CurrentPlayer == EPlayer.White) ? "White's turn" : "Black's turn";
					if (m_game.InCheck(m_game.CurrentPlayer))
					{
						lblStatus.Text += ": IN CHECK";
					}

					break;
				case Game.EState.NotStarted:
				default:
					lblStatus.Text = "Press 'New Game' to start";
					break;
			}
		}

		private void UpdateMoveCounter()
		{
			if (m_game.State != Game.EState.NotStarted)
			{
				lblMoves.Text = "Moves: " + m_game.MoveCount;
			}
			else
			{
				lblMoves.Text = "";
			}
		}

		private void DoFinished()
		{
			EnableBoardSquares(false);
			
			EWinner eWinner = m_game.Winner;
			switch (eWinner)
			{
				case EWinner.White:
					lblStatus.Text = "White wins!";
					MessageBox.Show(this, "White wins!");
					break;
				case EWinner.Black:
					lblStatus.Text = "Black wins!";
					MessageBox.Show(this, "Black wins!");
					break;
				case EWinner.Draw:
					lblStatus.Text = "It was a draw!";
					MessageBox.Show(this, "It was a draw!");
					break;
				case EWinner.Unfinished:
				default:
					break;
			}			
		}

		private void DoInput(Position pos)
		{
			switch (m_inputState)
			{
				case EInputState.SelectingMove:
					if (m_selectedPiece != null && m_selectedPiece.IsValidMove(pos))
					{
						SMove move;
						m_selectedPiece.GetMove(pos, out move); // no need to check return value since IsValidMove() above provides guarantee

						// Handle pawn promotion
						if (move.type == EMoveType.PawnPromotion)
						{
							PawnPromotionForm form = new PawnPromotionForm();
							form.ShowDialog();
							if (form.DialogResult == DialogResult.OK)
							{
								move.promotionType = form.PieceType;
							}	
						}						

						DoMove(move);
					}
					else
					{
						// Not a valid move, so see if this was an attempt to 
						// switch the currently selected piece instead
						goto case EInputState.SelectingPiece;
					}

					break;
				case EInputState.SelectingPiece:
					IPiece piece = m_game.GetPiece(pos);
					if (piece != null && piece.Owner == m_game.CurrentPlayer)
					{
						if (m_selectedPiece != null)
						{
							// We currently have a selected piece, so clear the valid move
							// colors for that piece
							ResetBoardColors();
						}

						DrawValidMoves(piece);
						m_selectedPiece = piece;
						m_inputState = EInputState.SelectingMove;
					}

					break;
				
				case EInputState.None:
				default:
					// Do nothing
					break;
			}
		}

		private void DrawValidMoves(IPiece piece)
		{
			// highlight square of selected piece
			Button button = m_buttons[piece.Position.row, piece.Position.col];
			button.BackColor = Color.SteelBlue;
			button.FlatAppearance.MouseOverBackColor = Color.SteelBlue;
			button.FlatAppearance.MouseDownBackColor = Color.SteelBlue;

			// highlight squares of all valid moves
			foreach (SMove move in piece.ValidMoves)
			{
				button = m_buttons[move.newPos.row, move.newPos.col];
				button.BackColor = Color.LightSteelBlue;
				button.FlatAppearance.MouseOverBackColor = Color.LightSteelBlue;
				button.FlatAppearance.MouseDownBackColor = Color.LightSteelBlue;
			}		
		}

		private void DoMove(SMove move)
		{
			m_inputState = EInputState.None;
			m_selectedPiece = null;
			m_game.RequestMove(move);
			ResetBoardColors();
			UpdateGame();
		}

		private void buttonNewGame_Click(object sender, EventArgs e)
		{
			Reset();
			StartGame();
		}

		private void b00_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 0));
		}

		private void b01_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 1));
		}

		private void b02_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 2));
		}

		private void b03_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 3));
		}

		private void b04_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 4));
		}

		private void b05_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 5));
		}

		private void b06_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 6));
		}

		private void b07_Click(object sender, EventArgs e)
		{
			DoInput(new Position(0, 7));
		}

		private void b10_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 0));
		}

		private void b11_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 1));
		}

		private void b12_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 2));
		}

		private void b13_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 3));
		}

		private void b14_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 4));
		}

		private void b15_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 5));
		}

		private void b16_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 6));
		}

		private void b17_Click(object sender, EventArgs e)
		{
			DoInput(new Position(1, 7));
		}

		private void b20_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 0));
		}

		private void b21_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 1));
		}

		private void b22_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 2));
		}

		private void b23_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 3));
		}

		private void b24_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 4));
		}

		private void b25_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 5));
		}

		private void b26_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 6));
		}

		private void b27_Click(object sender, EventArgs e)
		{
			DoInput(new Position(2, 7));
		}

		private void b30_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 0));
		}

		private void b31_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 1));
		}

		private void b32_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 2));
		}

		private void b33_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 3));
		}

		private void b34_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 4));
		}

		private void b35_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 5));
		}

		private void b36_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 6));
		}

		private void b37_Click(object sender, EventArgs e)
		{
			DoInput(new Position(3, 7));
		}

		private void b40_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 0));
		}

		private void b41_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 1));
		}

		private void b42_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 2));
		}

		private void b43_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 3));
		}

		private void b44_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 4));
		}

		private void b45_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 5));
		}

		private void b46_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 6));
		}

		private void b47_Click(object sender, EventArgs e)
		{
			DoInput(new Position(4, 7));
		}

		private void b50_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 0));
		}

		private void b51_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 1));
		}

		private void b52_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 2));
		}

		private void b53_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 3));
		}

		private void b54_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 4));
		}

		private void b55_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 5));
		}

		private void b56_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 6));
		}

		private void b57_Click(object sender, EventArgs e)
		{
			DoInput(new Position(5, 7));
		}

		private void b60_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 0));
		}

		private void b61_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 1));
		}

		private void b62_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 2));
		}

		private void b63_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 3));
		}

		private void b64_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 4));
		}

		private void b65_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 5));
		}

		private void b66_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 6));
		}

		private void b67_Click(object sender, EventArgs e)
		{
			DoInput(new Position(6, 7));
		}

		private void b70_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 0));
		}

		private void b71_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 1));
		}

		private void b72_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 2));
		}

		private void b73_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 3));
		}

		private void b74_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 4));
		}

		private void b75_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 5));
		}

		private void b76_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 6));
		}

		private void b77_Click(object sender, EventArgs e)
		{
			DoInput(new Position(7, 7));
		}


		private enum EInputState
		{
			None,
			SelectingPiece,
			SelectingMove
		}


		#region Members

		private Game m_game;
		private Button[,] m_buttons;
		private EInputState m_inputState;
		private IPiece m_selectedPiece;

		#endregion // Members
	}
}