﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Chessticles
{
	public enum EPieceType
	{
		None,
		Pawn,
		Knight,
		Rook,
		Bishop,
		Queen,
		King,
	}


	public abstract class Piece : IPiece
	{
		#region Properties

		public EPieceType Type
		{
			get { return m_eType; }
		}

		public EPlayer Owner
		{
			get { return m_eOwner; }
		}

		public Position Position
		{
			get { return m_position; }
			set
			{
				m_position = value;
				if (m_model.TestingForCheck == false)
				{
					++m_iMoveCount;
				}
			}
		}

		public bool OnBoard
		{
			get { return m_bOnBoard; }
			set
			{
				m_bOnBoard = value;
				if (!m_bOnBoard && m_model.TestingForCheck == false)
				{
					m_validMoves.Clear();
				}
			}
		}

		public int MoveCount
		{
			get { return m_iMoveCount; }
		}

		public List<SMove> ValidMoves
		{
			get { return m_validMoves; }
		}

		public List<SMove> TestMoves
		{
			get { return m_testMoves; }
		}

		#endregion // Properties


		public Piece(GameModel model, EPieceType type, EPlayer owner, Position pos)
		{
			m_model = model;
			m_eType = type;
			m_eOwner = owner;
			m_position = new Position(pos);
			m_bOnBoard = true;
			m_iMoveCount = 0;
			m_validMoves = new List<SMove>();
			m_testMoves = new List<SMove>();
		}

		public Piece(Piece piece)
		{
			m_model = piece.m_model;
			m_eType = piece.m_eType;
			m_eOwner = piece.m_eOwner;
			m_position = new Position(piece.m_position);
			m_bOnBoard = piece.m_bOnBoard;
			m_iMoveCount = piece.m_iMoveCount;
			m_validMoves = new List<SMove>();
			foreach (SMove move in piece.m_validMoves)
			{
				m_validMoves.Add(move);
			}
			m_testMoves = new List<SMove>();
			foreach (SMove move in piece.m_testMoves)
			{
				m_testMoves.Add(move);
			}
		}

		abstract public Piece Clone();
		abstract public void FindValidMoves(List<SMove> moves, int iDir);

		public bool IsValidMove(Position newPos)
		{
			if (m_bOnBoard)
			{
				foreach (SMove move in m_validMoves)
				{
					if (move.type != EMoveType.Invalid && move.newPos.Equals(newPos))
					{
						return true;
					}
				}
			}

			return false;
		}

		public bool GetMove(Position pos, out SMove outMove)
		{
			foreach (SMove move in m_validMoves)
			{
				if (move.newPos.Equals(pos))
				{
					outMove = move;
					return true;
				}
			}

			outMove = new SMove();
			return false;
		}

		// Updates moves, returns true if opponent is now in check
		public bool UpdateMoves(bool bTestOnly)
		{
			if (!m_bOnBoard)
			{
				return false;
			}

			List<SMove> moves = bTestOnly ? m_testMoves : m_validMoves;
			moves.Clear();
			
			int iDir = (m_eOwner == EPlayer.White) ? -1 : 1; // Forward direction for this piece

			FindValidMoves(moves, iDir);

			if (!bTestOnly)
			{
				PruneSelfCheckingMoves(moves);
			}

			return IsOpponentInCheck(moves);
		}

		protected void FindStraightLineMoves(Position delta, List<SMove> moves)
		{
			Position pos = m_position + delta;
			bool bHitAnotherPiece = false;
			EPlayer owner = EPlayer.None;
			while (!bHitAnotherPiece && pos.IsValid() && (owner = m_model.GetPositionOwner(pos)) != m_eOwner)
			{
				moves.Add(new SMove(this, new Position(pos.row, pos.col)));
				pos += delta;
				bHitAnotherPiece = (owner != EPlayer.None);
			}
		}

		protected bool IsOpponentInCheck(List<SMove> moves)
		{
			// Test if any moves put opponent in check
			EPlayer opponent = (m_eOwner == EPlayer.White) ? EPlayer.Black : EPlayer.White;
			foreach (SMove move in moves)
			{
				IPiece piece = m_model.GetPiece(move.newPos);
				if (piece != null && piece.Owner == opponent && piece.Type == EPieceType.King)
				{
					return true;
				}
			}

			return false;
		}

		protected void PruneSelfCheckingMoves(List<SMove> moves)
		{
			// Test if any moves put self in check, and remove them if so		
			for (int i = moves.Count - 1; i >= 0; --i)
			{
				if (m_model.TestIfMoveCausesSelfCheck(moves[i]) == true)
				{
					moves.RemoveAt(i);
				}
			}				
		}


		#region Members

		protected GameModel m_model;
		protected EPieceType m_eType;
		protected EPlayer m_eOwner;
		protected Position m_position;
		protected bool m_bOnBoard;
		protected int m_iMoveCount;
		protected List<SMove> m_validMoves;
		protected List<SMove> m_testMoves;

		#endregion // Members
	}

}
